from typing import Literal

import numpy as np
from csnlp.wrappers import Mpc
from mpcrl.agents.agent import Agent, SymType


class PwaAgent(Agent[SymType]):
    """An agent who has knowledge of it's own PWA dynamics and can use this to do things such as
    identify PWA regions given state and control trajectories."""

    buffer = 1e-7  # buffer for checking boundaries of PWA regions

    def __init__(
        self,
        mpc: Mpc,
        fixed_parameters: dict,
        pwa_system: dict,
        warmstart: Literal["last", "last-successful"] = "last-successful",
        name: str = None,
    ) -> None:
        """Initialise the agent.

        Parameters
        ----------
        mpc : Mpc[casadi.SX or MX]
            The MPC controller used as policy provider by this agent. The instance is
            modified in place to create the approximations of the state function `V(s)`
            and action value function `Q(s,a)`, so it is recommended not to modify it
            further after initialization of the agent. Moreover, some parameter and
            constraint names will need to be created, so an error is thrown if these
            names are already in use in the mpc. These names are under the attributes
            `perturbation_parameter`, `action_parameter` and `action_constraint`.
        fixed_parameters : dict[str, array_like] or collection of, optional
            A dict (or collection of dict, in case of `csnlp.MultistartNlp`) whose keys
            are the names of the MPC parameters and the values are their corresponding
            values. Use this to specify fixed parameters, that is, non-learnable. If
            `None`, then no fixed parameter is assumed.
        pwa_system: dict
            Contains {S, R, T, A, B, c, [Ac_j]}, where each is a list of matrices defining dynamics.
            When the inequality S[i]x + R[i]u <= T[i] is true, the dynamics are x^+ = A[i]x + B[i]u + c[i] + sum_j Ac_j[i] x_j.
        warmstart: 'last' or 'last-successful', optional
            The warmstart strategy for the MPC's NLP. If 'last-successful', the last
            successful solution is used to warm start the solver for the next iteration.
            If 'last', the last solution is used, regardless of success or failure.
        name : str, optional
            Name of the agent. If `None`, one is automatically created from a counter of
            the class' instancies."""
        super().__init__(mpc, fixed_parameters, warmstart, name)
        self.S = pwa_system["S"]
        self.R = pwa_system["R"]
        self.T = pwa_system["T"]
        self.A = pwa_system["A"]
        self.B = pwa_system["B"]
        self.c = pwa_system["c"]
        self.Ac = pwa_system["Ac"]

        self.num_neighbours = len(self.Ac[0])

    def next_state(self, x: np.ndarray, u: np.ndarray, xc: list[np.ndarray] = None):
        """Increment the dynamics as x+ = A[i]x + B[i]u + c[i] + sum_j Ac[i]_j xc_j
        if S[i]x + R[i]u <= T.
        If the coupled states xc are not passed the coupling part of dynamics is ignored.
        """
        for i in range(len(self.S)):
            if all(self.S[i] @ x + self.R[i] @ u <= self.T[i] + self.buffer):
                if xc is None:
                    return self.A[i] @ x + self.B[i] @ u + self.c[i]
                else:
                    return (
                        self.A[i] @ x
                        + self.B[i] @ u
                        + self.c[i]
                        + sum(self.Ac[i][j] @ xc[j] for j in range(self.num_neighbours))
                    )

        raise RuntimeError("Didn't find PWA region for given state-control.")

    def eval_sequences(
        self, x0: np.ndarray, u: np.ndarray, xc: list[np.ndarray] = None
    ):
        """Evaluate all possible sqitching sequences of PWA dynamics by rolling out
        dynamics from state x, applying control u, and coupled states xc."""

        N = u.shape[1]  # horizon length
        s = [[0] * N]  # list of sequences, start with just zeros
        x = x0  # temp state

        for k in range(N):
            current_regions = self.identify_regions(x, u[:, [k]])
            for i in range(len(current_regions)):
                if i == 0:  # first one gets appended to all current sequences
                    for j in range(len(s)):
                        s[j][k] = current_regions[i]
                else:
                    # for other identified regions, they define new sequences
                    s_temp = []
                    for j in range(len(s)):
                        s_temp.append(s[j].copy())
                        s_temp[j][k] = current_regions[i]
                    s = s + s_temp
            # rollout dynamics by one
            if xc is None:
                x = self.next_state(x, u[:, [k]])
            else:
                x = self.next_state(
                    x, u[:, [k]], [xc[i][:, [k]] for i in range(self.num_neighbours)]
                )
        return s

    def identify_regions(self, x: np.ndarray, u: np.ndarray, eps: float = 0.0000001):
        """Generate the indices of the regions where Sx+Ru<=T + eps is true."""

        regions = []
        for i in range(len(self.S)):
            if all(self.S[i] @ x + self.R[i] @ u <= self.T[i] + eps):
                regions.append(i)
        return regions

    def set_sequence(self, s: list[int]):
        """Modify the parameters in the constraints of the ADMM Mpc to
        enforce the sequence s"""

        # TODO confirm that these parameters have been named correctly in the MPC
        #
        for i in range(len(s)):
            self.fixed_parameters[f"A_{i}"] = self.A[s[i]]
            self.fixed_parameters[f"B_{i}"] = self.B[s[i]]
            self.fixed_parameters[f"c_{i}"] = self.c[s[i]]
            self.fixed_parameters[f"S_{i}"] = self.S[s[i]]
            self.fixed_parameters[f"R_{i}"] = self.R[s[i]]
            self.fixed_parameters[f"T_{i}"] = self.T[s[i]]
            for j in range(self.num_neighbours):
                self.fixed_parameters[f"Ac_{i}_{j}"] = self.Ac[s[i]][j]



from typing import Any

import casadi as cs
import numpy as np
from csnlp.util.io import SupportsDeepcopyAndPickle
from gymnasium import Env
from mpcrl import Agent
from mpcrl.core.callbacks import AgentCallbackMixin
from mpcrl.core.exploration import ExplorationStrategy, NoExploration
from mpcrl.util.named import Named

from dmpcpwa.mpc.mpc_mld import MpcMld


class MldAgent(Agent):
    """A pwa agent who uses an mld controller."""

    def __init__(
        self,
        mpc: MpcMld,
    ) -> None:
        """Constructor is overriden and the super class' instructor is not called as
        this agent uses an mpc that does not inheret from the MPC baseclass."""

        # this below is just to keep this class compatible with Agent init
        Named.__init__(self, None)
        SupportsDeepcopyAndPickle.__init__(self)
        AgentCallbackMixin.__init__(self)
        self._fixed_pars = None
        self._exploration: ExplorationStrategy = NoExploration()
        self._store_last_successful = True
        self._last_action_on_fail = False
        self._last_solution = None
        self._last_action = None

        self.mpc = mpc
        self.x_pred = None  # stores most recent predicted state after being solved
        self.u_pred = None  # stores most recent predicted control after being solved
        self.cost_pred = None  # stores most recent predicted cost after being solved
        self.run_time = None  # stores most recent solve time of mpc
        self.node_count = None  # stores the node count from last MPC solution
        self.num_bin_vars = (
            None  # stors the number of binary variables in the model AFTER the presolve
        )

    def evaluate(
        self,
        env: Env,
        episodes: int,
        deterministic: bool = True,
        seed: int = None,
        raises: bool = True,
        env_reset_options: dict[str, Any] = None,
    ):
        """Evaluates the agent in a given environment. Overriding the function of Agent
        to use the mld_mpc instead.

        Parameters
        ----------
        env : Env[ObsType, ActType]
            A gym environment where to test the agent in.
        episodes : int
            Number of evaluation episodes.
        deterministic : bool, optional
            Whether the agent should act deterministically; by default, `True`.
        seed : None, int or sequence of ints, optional
            Agent's and each env's RNG seed.
        raises : bool, optional
            If `True`, when any of the MPC solver runs fails, or when an update fails,
            the corresponding error is raised; otherwise, only a warning is raised.
        env_reset_options : dict, optional
            Additional information to specify how the environment is reset at each
            evalution episode (optional, depending on the specific environment).

        Returns
        -------
        array of doubles
            The cumulative returns (one return per evaluation episode)."""
        returns = np.zeros(episodes)
        self.on_validation_start(env)
        seeds = map(int, np.random.SeedSequence(seed).generate_state(episodes))

        for episode, current_seed in zip(range(episodes), seeds):
            self.reset(current_seed)
            state, _ = env.reset(seed=current_seed, options=env_reset_options)
            truncated, terminated, timestep = False, False, 0
            self.on_episode_start(env, episode, state)

            while not (truncated or terminated):
                # changed origonal agents evaluate here to use the mld mpc
                action = self.get_control(state)
                action = cs.DM(action)

                state, r, truncated, terminated, _ = env.step(action)
                self.on_env_step(env, episode, timestep)

                returns[episode] += r
                timestep += 1
                self.on_timestep_end(env, episode, timestep)

            self.on_episode_end(env, episode, returns[episode])

        self.on_validation_end(env, returns)
        return returns

    def get_control(self, state):
        u, info = self.mpc.solve_mpc(state)
        self.x_pred = info["x"]
        self.u_pred = info["u"]
        self.cost_pred = info["cost"]
        self.run_time = info["run_time"]
        self.node_count = info["nodes"]
        self.num_bin_vars = info["bin_vars"]
        return u

    def set_cost(self, Q_x, Q_u, x_goal: np.ndarray = None, u_goal: np.ndarray = None):
        """Set cost of the agents mpc-MIP as sum_k x(k)' * Q_x * x(k) + u(k)' * Q_u * u(k).
        Restricted to quadratic in the states and control.
        If x_goal or u_goal passed the cost uses (x-x_goal) and (u_goal)"""

        self.mpc.set_cost(Q_x, Q_u, x_goal, u_goal)

    def get_predicted_state(self, shifted=False) -> np.ndarray:
        """Returns the predicted state trajectory from the most recent solve of the local MPC.
        If shifted is true, the sequence is shifted by one, with the final time step duplicated.
        If no mpc has been solved, the returned value will be none."""
        if shifted and self.x_pred is not None:
            x_pred_shifted = self.x_pred.copy()
            return np.concatenate(
                (x_pred_shifted[:, 1:], x_pred_shifted[:, -1:]), axis=1
            )
        else:
            return self.x_pred

    def get_predicted_cost(self):
        """Return cost of most recent MPC solution."""
        return self.cost_pred



import logging
from typing import Any, Literal

import casadi as cs
import numpy as np
from dmpcrl.core.admm import AdmmCoordinator
from dmpcrl.mpc.mpc_admm import MpcAdmm
from gymnasium import Env
from mpcrl import Agent
from mpcrl.agents.agent import ActType, ObsType

from dmpcpwa.agents.pwa_agent import PwaAgent

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
console_handler = logging.StreamHandler()
console_handler.setLevel(logging.DEBUG)
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
console_handler.setFormatter(formatter)
logger.addHandler(console_handler)


class SqpAdmmCoordinator(Agent):
    """Coordinates the SQP-ADMM algorithm for PWA agents"""

    admm_iters = 50
    eps = 1e-6  # threshold for comparing differences in costs

    def __init__(
        self,
        local_mpcs: list[MpcAdmm],
        local_fixed_parameters: list[dict],
        systems: list[dict],
        G: list[list[int]],
        Adj: np.ndarray,
        rho: float,
        warmstart: Literal["last", "last-successful"] = "last-successful",
        name: str = None,
    ) -> None:
        """Instantiates the coordinator, creating n PWA agents.

        Parameters
        ----------
        local_mpcs: list[MpcAdmm[SymType]]
            List of local MPCs for agents.
        local_fixed_parameters: List[dict]
            List of dictionaries for fixed parameters for agents.
        systems : dict
            PWA model for each agent.
        G :  List[List[int]]
            Map of local to global vars in ADMM.
        Adj: np.ndarray
            Adjacency matrix for agent coupling
        rho: float
            Augmented lagrangian penalty term."""

        # to the super class we pass the first local mpc just to satisfy the constructor
        # we copy it so the parameters don't double up etc.
        super().__init__(
            local_mpcs[0].copy(), local_fixed_parameters[0].copy(), warmstart, name
        )

        # construct the agents
        self.n = len(local_mpcs)
        self.agents: list[PwaAgent] = []
        for i in range(self.n):
            self.agents.append(
                PwaAgent(local_mpcs[i], local_fixed_parameters[i], systems[i])
            )

        # create ADMM coordinator
        self.N = local_mpcs[0].horizon
        self.Adj = Adj
        self.nx_l = local_mpcs[0].nx_l
        self.nu_l = local_mpcs[0].nu_l

        # coordinator of ADMM using 1 iteration as g_admm coordinator checks sequences every ADMM iter
        self.admm_coordinator = AdmmCoordinator(
            self.agents,
            G,
            self.N,
            self.nx_l,
            self.nu_l,
            rho,
            iters=self.admm_iters,
        )

    def evaluate(
        self,
        env: Env[ObsType, ActType],
        episodes: int,
        deterministic: bool = True,
        seed: int = None,
        raises: bool = True,
        env_reset_options: dict[str, Any] = None,
    ):
        returns = np.zeros(episodes)
        self.on_validation_start(env)
        seeds = map(int, np.random.SeedSequence(seed).generate_state(episodes))

        for episode, current_seed in zip(range(episodes), seeds):
            for agent in self.agents:
                agent.reset(current_seed)
            state, _ = env.reset(seed=current_seed, options=env_reset_options)
            truncated, terminated, timestep = False, False, 0
            for agent in self.agents:
                agent.on_episode_start(env, episode)

            while not (truncated or terminated):
                action, sol_list = self.sqp_admm_control(state, deterministic)
                for i in range(len(sol_list)):
                    if not sol_list[i].success:
                        self.agents[i].on_mpc_failure(
                            episode, timestep, sol_list[i].status, raises
                        )

                state, r, truncated, terminated, _ = env.step(action)
                for agent in self.agents:
                    agent.on_env_step(env, episode, timestep)

                returns[episode] += r
                timestep += 1
                for agent in self.agents:
                    agent.on_timestep_end(env, episode, timestep)

            for agent in self.agents:
                agent.on_episode_end(env, episode, returns[episode])

        for agent in self.agents:
            agent.on_validation_end(env, returns)
        return returns

    def sqp_admm_control(self, state, deterministic):
        # break global state into local pieces
        x = [state[self.nx_l * i : self.nx_l * (i + 1), :] for i in range(self.n)]

        xc = [None] * self.n

        # initial feasible control guess
        u = [np.zeros((self.nu_l, self.N)) for i in range(self.n)]

        # generate initial feasible coupling via dynamics rollout
        x_rout = self.dynamics_rollout(x, u)

        continue_flag = True
        first_iter_flag = True

        logger.debug("new sqp iter")

        while continue_flag:
            continue_flag = False

            seqs = [None] * self.n

            # generate sequences for all agents
            for i in range(self.n):
                if first_iter_flag:
                    seqs[i] = self.agents[i].eval_sequences(
                        x[i],
                        u[i],
                        [x_rout[j] for j in range(self.n) if self.Adj[i, j] == 1],
                    )

                    # first iter flag changed below

                else:
                    seqs[i] = self.agents[i].eval_sequences(
                        x[i],
                        u[i],
                        xc[i],  # use local ADMM vars if not first iter
                    )

            logger.debug(f"Sequences: {seqs}")

            # set up coordinated enumeration of sequences

            # contains the numbers of sequences to enumerate before agents should switch sequence
            cue_list = [None] * self.n
            # contains the current indexes for sequences
            index_list = [None] * self.n
            for i in range(self.n):
                index_list[i] = 0
                temp_val = 1
                for j in range(i):
                    temp_val *= len(seqs[j])
                cue_list[i] = temp_val

            num_seqs = cue_list[-1] * len(seqs[-1])
            counter = 1

            while counter <= num_seqs:
                for i in range(self.n):
                    self.agents[i].set_sequence(seqs[i][index_list[i]])

                # solve current sequence choice
                action_list, sol_list, error_flag = self.admm_coordinator.solve_admm(
                    state
                )
                cost_temp = sum(sol_list[i].f for i in range(self.n))
                if first_iter_flag or cost_opt - cost_temp > self.eps:
                    if not first_iter_flag:
                        logger.debug(f"cost decreased by: {cost_opt - cost_temp}")

                    # extract vars over horizon
                    for i in range(self.n):
                        u[i] = np.asarray(sol_list[i].vals["u"])
                        xc_out = np.asarray(sol_list[i].vals["x_c"])
                        xc_temp = []
                        for j in range(self.agents[i].num_neighbours):
                            xc_temp.append(
                                xc_out[self.nx_l * j : self.nx_l * (j + 1), :]
                            )
                        xc[i] = xc_temp
                        action_list_opt = action_list

                    if first_iter_flag:
                        first_iter_flag = False

                    continue_flag = True

                    # break the loop at first improvement (taking first descent direction)
                    break

                # increment the enumeration of sequences
                for i in range(self.n):
                    if counter % cue_list[i] == 0:
                        index_list[i] = (index_list[i] + 1) % len(seqs[i])

                counter += 1

        return cs.DM(action_list_opt), sol_list

    def dynamics_rollout(self, x: list[np.ndarray], u: list[np.ndarray]):
        """For a given state and u, rollout the agents' dynamics step by step."""
        x_temp = [np.zeros((self.nx_l, self.N)) for i in range(self.n)]

        for i in range(self.n):
            x_temp[i][:, [0]] = x[i]  # add the first known states to the temp
        for k in range(1, self.N):
            for i in range(self.n):
                xc_temp = []
                for j in range(self.n):
                    if self.Adj[i, j] == 1:
                        xc_temp.append(x_temp[j][:, [k - 1]])
                x_temp[i][:, [k]] = self.agents[i].next_state(
                    x_temp[i][:, [k - 1]], u[i][:, [k]], xc_temp
                )
        return x_temp



import logging

import gurobipy as gp
import numpy as np

from dmpcpwa.mpc.mpc_mld import MpcMld

logger = logging.getLogger(__name__)
logger.setLevel(logging.CRITICAL)
console_handler = logging.StreamHandler()
console_handler.setLevel(logging.CRITICAL)
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
console_handler.setFormatter(formatter)
logger.addHandler(console_handler)


class MpcMldCentDecup(MpcMld):
    """ "An mpc that converts the networked PWA mpc problem to MLD form.
    The PWA systems are assumed decoupled in the dynamics."""

    def __init__(self, systems: list[dict], n: int, N: int, verbose=False) -> None:
        """Instantiate the mpc.

        Parameters
        ----------
        system: List[dict]
            List of dictionary which contain the definitions of the PWA systems {S, R, T, A, B, c, D, E, F, G}.
             When S[i]x+R[x]u <= T[i] -> x+ = A[i]x + B[i]u + c[i].
             For MLD conversion the state and input must be constrained: Dx <= E, Fu <= G.
        n: int
            The number of agents.
        N:int
            Prediction horizon length."""

        nx_l = systems[0]["A"][0].shape[0]
        nu_l = systems[0]["B"][0].shape[1]

        # build mld model

        mpc_model = gp.Model("mld_mpc")
        mpc_model.setParam("OutputFlag", verbose)
        mpc_model.setParam("Heuristics", 0)
        # mpc_model.setParam("MIPStart", 1)  # using warm-starting from previous sol

        # Uncomment if you need to differentiate between infeasbile and unbounded
        mpc_model.setParam("DualReductions", 0)

        x = mpc_model.addMVar(
            (n * nx_l, N + 1), lb=-float("inf"), ub=float("inf"), name="x"
        )  # state
        u = mpc_model.addMVar(
            (n * nu_l, N), lb=-float("inf"), ub=float("inf"), name="u"
        )  # control

        for i in range(n):
            # pull out local parts for agent i
            x_l = x[nx_l * i : nx_l * (i + 1), :]
            u_l = u[nu_l * i : nu_l * (i + 1), :]

            self.create_MLD_dynamics_and_constraints(systems[i], mpc_model, x_l, u_l, N)

        # IC constraint - gets updated everytime solve_mpc is called
        self.IC = mpc_model.addConstr(x[:, [0]] == np.zeros((n * nx_l, 1)), name="IC")

        # assign parts of model to be used by class later
        self.mpc_model = mpc_model
        self.x = x
        self.u = u
        # self.n = n * nx_l
        self.m = n * nu_l
        self.N = N

        logger.critical("MLD MPC setup complete.")



import logging

import gurobipy as gp
import numpy as np

logger = logging.getLogger(__name__)
logger.setLevel(logging.CRITICAL)
console_handler = logging.StreamHandler()
console_handler.setLevel(logging.CRITICAL)
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
console_handler.setFormatter(formatter)
logger.addHandler(console_handler)


class MpcMld:
    """An MPC that converts a PWA mpc problem into a MIP."""

    def __init__(self, system: dict, N: int, verbose=False) -> None:
        """Instantiate the mld based mpc. In the constructor pwa system is converted
        to mld and the associated dynamics and constraints are created, along with states
        and control variables.

        Parameters
        ----------
        system: dict
            Dictionary containing the definition of the PWA system {S, R, T, A, B, c, D, E, F, G}.
             When S[i]x+R[x]u <= T[i] -> x+ = A[i]x + B[i]u + c[i].
             For MLD conversion the state and input must be constrained: Dx <= E, Fu <= G.
        N:int
            Prediction horizon length."""

        n = system["A"][0].shape[0]
        m = system["B"][0].shape[1]

        # build mld model

        mpc_model = gp.Model("mld_mpc")
        mpc_model.setParam("OutputFlag", verbose)
        mpc_model.setParam("Heuristics", 0)
        # mpc_model.setParam("MIPStart", 1)  # using warm-starting from previous sol

        # Uncomment if you need to differentiate between infeasbile and unbounded
        mpc_model.setParam("DualReductions", 0)

        x = mpc_model.addMVar(
            (n, N + 1), lb=-float("inf"), ub=float("inf"), name="x"
        )  # state
        u = mpc_model.addMVar(
            (m, N), lb=-float("inf"), ub=float("inf"), name="u"
        )  # control

        # create MLD dynamics from PWA
        self.create_MLD_dynamics_and_constraints(system, mpc_model, x, u, N)

        # IC constraint - gets updated everytime solve_mpc is called
        self.IC = mpc_model.addConstr(x[:, [0]] == np.zeros((n, 1)), name="IC")

        # assign parts of model to be used by class later
        self.mpc_model = mpc_model
        self.x = x
        self.u = u
        self.n = n
        self.m = m
        self.N = N

        logger.critical("MLD MPC setup complete.")

    def create_MLD_dynamics_and_constraints(self, system, mpc_model, x, u, N):
        # extract values from system
        S = system["S"]
        R = system["R"]
        T = system["T"]
        A = system["A"]
        B = system["B"]
        c = system["c"]
        D = system["D"]
        E = system["E"]
        F = system["F"]
        G = system["G"]
        s = len(S)  # number of PWA regions
        n = A[0].shape[0]
        m = B[0].shape[1]

        M_st = [None] * s  # The upper bound for each region
        model_lin = gp.Model("linear model for mld set-up")
        model_lin.setParam("OutputFlag", 0)

        x_temp = model_lin.addMVar(
            (n, 1), lb=-float("inf"), ub=float("inf"), name="x_lin"
        )
        u_temp = model_lin.addMVar(
            (m, 1), lb=-float("inf"), ub=float("inf"), name="u_lin"
        )
        model_lin.addConstr(D @ x_temp <= E, name="state constraints")
        model_lin.addConstr(F @ u_temp <= G, name="control constraints")
        for i in range(s):
            obj = S[i] @ x_temp + R[i] @ u_temp - T[i]
            M_st[i] = np.zeros(obj.shape)
            for j in range(obj.shape[0]):
                model_lin.setObjective(obj[j, 0], gp.GRB.MAXIMIZE)
                model_lin.update()
                model_lin.optimize()
                M_st[i][j, 0] = model_lin.ObjVal
        logger.critical(
            "Solved linear model for PWA region bounds, M_star = " + str(M_st)
        )

        # bounds for state updates

        M_ub = np.zeros((n, 1))
        m_lb = np.zeros((n, 1))
        for j in range(n):
            M_regions = [None] * s
            m_regions = [None] * s
            for i in range(s):
                obj = A[i][j, :] @ x_temp + B[i][j, :] @ u_temp + c[i][j, :]
                model_lin.setObjective(obj, gp.GRB.MAXIMIZE)
                model_lin.update()
                model_lin.optimize()
                M_regions[i] = model_lin.ObjVal
                model_lin.setObjective(obj, gp.GRB.MINIMIZE)
                model_lin.update()
                model_lin.optimize()
                m_regions[i] = model_lin.ObjVal
            M_ub[j] = np.max(M_regions)
            m_lb[j] = np.min(m_regions)
        logger.critical(
            "Solved linear model for PWA state update bounds, M = "
            + str(M_ub.T)
            + "', m = "
            + str(m_lb.T)
            + "'"
        )

        # auxillary var z has 3 dimensions. (Region, state, time)
        z = mpc_model.addMVar((s, n, N), lb=-float("inf"), ub=float("inf"), name="z")
        # binary auxillary var
        delta = mpc_model.addMVar((s, N), vtype=gp.GRB.BINARY, name="delta")

        # constraint that only 1 delta can be active at each time step
        mpc_model.addConstrs(
            (gp.quicksum(delta[i, j] for i in range(s)) == 1 for j in range(N)),
            name="Delta sum constraints",
        )

        # constraints along predictions horizon for dynamics, state and control
        for k in range(N):
            # Set the branch priority of the binaries proportional to earlyness (earlier is more important)
            for i in range(s):
                # delta[i, k].setAttr('BranchPriority', s*N - s*k - i)
                # delta[i, k].setAttr('BranchPriority', N-k)
                pass

            # add state and input constraints to model, then binary and auxillary constraint, then dynamics constraints

            mpc_model.addConstr(D @ x[:, [k]] <= E, name="state constraints")
            mpc_model.addConstr(F @ u[:, [k]] <= G, name=f"control constraints_{k}")

            mpc_model.addConstrs(
                (
                    S[i] @ x[:, [k]] + R[i] @ u[:, [k]] - T[i]
                    <= M_st[i] * (1 - delta[i, [k]])
                    for i in range(s)
                ),
                name="Region constraints",
            )

            mpc_model.addConstrs(
                (z[i, :, k] <= M_ub @ (delta[i, [k]]) for i in range(s)),
                name="Z leq binary constraints",
            )

            mpc_model.addConstrs(
                (z[i, :, k] >= m_lb @ (delta[i, [k]]) for i in range(s)),
                name="Z geq binary constraints",
            )

            mpc_model.addConstrs(
                (
                    z[i, :, k].reshape(n, 1)
                    <= A[i] @ x[:, [k]]
                    + B[i] @ u[:, [k]]
                    + c[i]
                    - (m_lb * (1 - delta[i, [k]]))
                    for i in range(s)
                ),
                name="Z leq state constraints",
            )

            mpc_model.addConstrs(
                (
                    z[i, :, k].reshape(n, 1)
                    >= (A[i] @ x[:, [k]])
                    + (B[i] @ u[:, [k]])
                    + c[i]
                    - (M_ub * (1 - delta[i, [k]]))
                    for i in range(s)
                ),
                name="Z geq state constraints",
            )

            mpc_model.addConstr(
                x[:, [k + 1]]
                == gp.quicksum(z[i, :, k].reshape(n, 1) for i in range(s)),
                name="dynamics",
            )

        mpc_model.addConstr(
            D @ x[:, [N]] <= E, name="state constraints"
        )  # final state constraint

        # trivial terminal constraint condition x(N) = 0
        # mpc_model.addConstr(x[:, [N]] == np.zeros((n, 1)))

    def set_cost(self, Q_x, Q_u, x_goal: np.ndarray = None, u_goal: np.ndarray = None):
        """Set cost of the MIP as sum_k x(k)' * Q_x * x(k) + u(k)' * Q_u * u(k).
        Restricted to quadratic in the states and control.
        If x_goal or u_goal passed the cost uses (x-x_goal) and (u_goal)"""

        # construct zero goal points if not passed
        if x_goal is None:
            x_goal = np.zeros((self.x[:, [0]].shape[0], self.N + 1))
        if u_goal is None:
            u_goal = np.zeros((self.u[:, [0]].shape[0], self.N + 1))

        obj = 0
        for k in range(self.N):
            obj += (self.x[:, k] - x_goal[:, k].T) @ Q_x @ (
                self.x[:, [k]] - x_goal[:, [k]]
            ) + (self.u[:, k] - u_goal[:, k].T) @ Q_u @ (
                self.u[:, [k]] - u_goal[:, [k]]
            )
        obj += (
            (self.x[:, self.N] - x_goal[:, self.N].T)
            @ Q_x
            @ (self.x[:, [self.N]] - x_goal[:, [self.N]])
        )
        self.mpc_model.setObjective(obj, gp.GRB.MINIMIZE)

    def min_1_norm(self, x, Q=None):
        """returns a dummy variable t which should be minimised in the cost
        to substitute the minimization of |Qx| (1-norm). This functions will
        also add the required constraints on x to make the minimization of
        t equivalent to minimizing |Qx|. If Q not passed it is identity."""
        if x.shape[1] > 1:
            raise RuntimeError("x must be a columb vector.")
        if Q is None:
            Q = np.eye(x.shape[0])
        n = x.shape[0]
        t = self.mpc_model.addMVar((1, 1), lb=-float("inf"), ub=float("inf"), name="t")
        y = self.mpc_model.addMVar(
            x.shape, lb=-float("inf"), ub=float("inf"), name="y_dumb"
        )
        x = Q @ x
        for i in range(n):
            self.mpc_model.addConstr(x[i, :] <= y[i, :], name="dumb_leq")
            self.mpc_model.addConstr(-x[i, :] <= y[i, :], name="dumb_geq")
        self.mpc_model.addConstr(sum(y[i, :] for i in range(n)) == t, name="dumb_eq")
        return t

    def min_2_norm(self, x, Q):
        """return the term x'@Q@x."""
        # I have to do the tranpose part manually because gurobi does not support
        # taking the transpose of an Expr
        if x.shape[1] > 1:
            raise RuntimeError("x must be a columb vector.")
        n = x.shape[0]
        M = Q @ x
        return sum(x[i] * M[i] for i in range(n))

    def solve_mpc(self, state):
        self.IC.RHS = state
        self.mpc_model.optimize()
        if self.mpc_model.Status == 2:  # check for successful solve
            u = self.u.X
            x = self.x.X
            cost = self.mpc_model.objVal
        else:
            u = np.zeros((self.m, self.N))
            x = np.zeros(self.x.shape)
            cost = float("inf")
            logger.info("Infeasible")

        run_time = self.mpc_model.Runtime
        nodes = self.mpc_model.NodeCount
        try:
            bin_vars = self.mpc_model.presolve().NumBinVars
        except:
            bin_vars = 0

        self.mpc_model.NodeCount
        return u[:, [0]], {
            "x": x,
            "u": u,
            "cost": cost,
            "run_time": run_time,
            "nodes": nodes,
            "bin_vars": bin_vars,
        }



from typing import Any

import casadi as cs
import numpy as np
from gymnasium import Env
from mpcrl import Agent
from mpcrl.core.exploration import ExplorationStrategy, NoExploration


class NoControlAgent(Agent):
    """An agent who uses zero control input."""

    def __init__(self, nu: int) -> None:
        """Initialise the no control agent with control dimension nu."""
        self._exploration: ExplorationStrategy = NoExploration()  # to keep compatable
        self.nu = nu

    def evaluate(
        self,
        env: Env,
        episodes: int,
        deterministic: bool = True,
        seed: int = None,
        raises: bool = True,
        env_reset_options: dict[str, Any] = None,
    ):
        """Evaluates the agent in a given environment. Overriding the evaluation to not
        use control.

        Parameters
        ----------
        env : Env[ObsType, ActType]
            A gym environment where to test the agent in.
        episodes : int
            Number of evaluation episodes.
        deterministic : bool, optional
            Whether the agent should act deterministically; by default, `True`.
        seed : None, int or sequence of ints, optional
            Agent's and each env's RNG seed.
        raises : bool, optional
            If `True`, when any of the MPC solver runs fails, or when an update fails,
            the corresponding error is raised; otherwise, only a warning is raised.
        env_reset_options : dict, optional
            Additional information to specify how the environment is reset at each
            evalution episode (optional, depending on the specific environment).

        Returns
        -------
        array of doubles
            The cumulative returns (one return per evaluation episode)."""
        returns = np.zeros(episodes)
        self.on_validation_start(env)
        seeds = map(int, np.random.SeedSequence(seed).generate_state(episodes))

        for episode, current_seed in zip(range(episodes), seeds):
            self.reset(current_seed)
            state, _ = env.reset(seed=current_seed, options=env_reset_options)
            truncated, terminated, timestep = False, False, 0
            self.on_episode_start(env, episode)

            while not (truncated or terminated):
                # changed origonal agents evaluate here not control
                action = np.zeros((self.nu, 1))
                action = cs.DM(action)

                state, r, truncated, terminated, _ = env.step(action)
                self.on_env_step(env, episode, timestep)

                returns[episode] += r
                timestep += 1
                self.on_timestep_end(env, episode, timestep)

            self.on_episode_end(env, episode, returns[episode])

        self.on_validation_end(env, returns)
        return returns



import logging
from typing import Any, Literal

import casadi as cs
import matplotlib.pyplot as plt
import numpy as np
from dmpcrl.core.admm import AdmmCoordinator
from dmpcrl.mpc.mpc_admm import MpcAdmm
from gymnasium import Env
from mpcrl import Agent
from mpcrl.agents.agent import ActType, ObsType

from dmpcpwa.agents.pwa_agent import PwaAgent

ADMM_DEBUG_PLOT = False
DEBUG_PRINT = False

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
console_handler = logging.StreamHandler()
console_handler.setLevel(logging.DEBUG)
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
console_handler.setFormatter(formatter)
logger.addHandler(console_handler)


class GAdmmCoordinator(Agent):
    """Coordinates the greedy ADMM algorithm for PWA agents"""

    admm_iters = 50
    switching_iters = 50

    def __init__(
        self,
        local_mpcs: list[MpcAdmm],
        local_fixed_parameters: list[dict],
        systems: list[dict],
        G: list[list[int]],
        Adj: np.ndarray,
        rho: float,
        agent_class=PwaAgent,
        warmstart: Literal["last", "last-successful"] = "last-successful",
        name: str = None,
    ) -> None:
        """Instantiates the coordinator, creating n PWA agents.

        Parameters
        ----------
        local_mpcs: list[MpcAdmm[SymType]]
            List of local MPCs for agents.
        local_fixed_parameters: List[dict]
            List of dictionaries for fixed parameters for agents.
        systems : dict
            PWA model for each agent.
        G :  List[List[int]]
            Map of local to global vars in ADMM.
        Adj: np.ndarray
            Adjacency matrix for agent coupling
        rho: float
            Augmented lagrangian penalty term."""

        # to the super class we pass the first local mpc just to satisfy the constructor
        # we copy it so the parameters don't double up etc.
        super().__init__(
            local_mpcs[0].copy(), local_fixed_parameters[0].copy(), warmstart, name
        )

        # construct the agents
        self.n = len(local_mpcs)
        self.agents: list[PwaAgent] = []
        for i in range(self.n):
            self.agents.append(
                agent_class(local_mpcs[i], local_fixed_parameters[i], systems[i])
            )

        # create ADMM coordinator
        self.N = local_mpcs[0].horizon
        self.Adj = Adj
        self.nx_l = local_mpcs[0].nx_l
        self.nu_l = local_mpcs[0].nu_l

        self.prev_sol = None
        self.prev_traj = None

        # coordinator of ADMM using 1 iteration as g_admm coordinator checks sequences every ADMM iter
        self.admm_coordinator = AdmmCoordinator(
            self.agents,
            G,
            self.N,
            self.nx_l,
            self.nu_l,
            rho,
            iters=1,
        )

    def evaluate(
        self,
        env: Env[ObsType, ActType],
        episodes: int,
        deterministic: bool = True,
        seed: int = None,
        raises: bool = True,
        env_reset_options: dict[str, Any] = None,
    ):
        returns = np.zeros(episodes)
        self.on_validation_start(env)
        seeds = map(int, np.random.SeedSequence(seed).generate_state(episodes))

        for episode, current_seed in zip(range(episodes), seeds):
            self.reset(current_seed)
            for agent in self.agents:
                agent.reset(current_seed)
            state, _ = env.reset(seed=current_seed, options=env_reset_options)
            truncated, terminated, timestep = False, False, 0

            self.on_episode_start(env, episode, state)
            for agent in self.agents:
                agent.on_episode_start(env, episode, state)

            while not (truncated or terminated):
                action, _, _, _ = self.g_admm_control(state)

                state, r, truncated, terminated, _ = env.step(action)

                self.on_env_step(env, episode, timestep)
                for agent in self.agents:
                    agent.on_env_step(env, episode, timestep)

                returns[episode] += r
                timestep += 1

                self.on_timestep_end(env, episode, timestep)
                for agent in self.agents:
                    agent.on_timestep_end(env, episode, timestep)

            self.on_episode_end(env, episode, returns[episode])
            for agent in self.agents:
                agent.on_episode_end(env, episode, returns[episode])

        self.on_validation_end(env, returns)
        for agent in self.agents:
            agent.on_validation_end(env, returns)
        return returns

    def g_admm_control(self, state, warm_start=None):
        """Get the control for the given state. Warm start parameter is an initial guess for control actions."""
        seqs = [[0] * self.N for i in range(self.n)]  # switching seqs for agents

        xc = [None] * self.n
        x_pred = [None] * self.n
        action_list = None
        sol_list = None
        error_flag = False

        # break global state into local pieces
        x = [state[self.nx_l * i : self.nx_l * (i + 1), :] for i in range(self.n)]

        if warm_start is None:
            warm_start = [np.zeros((self.nu_l, self.N)) for i in range(self.n)]

        infeas_flag = False
        u = warm_start
        # generate initial feasible coupling via dynamics rollout
        try:
            x_rout = self.dynamics_rollout(x, u)
        except:
            logger.debug(f"Rollout of initial control guess {u} was infeasible.")
            infeas_flag = True

        if ADMM_DEBUG_PLOT:  # store control at each iter to plot ADMM convergence
            u_plot_list = [
                [np.zeros((self.nu_l, self.N)) for k in range(self.admm_iters)]
                for i in range(self.n)
            ]
            switch_plot_list = [[] for i in range(self.n)]
            z_plot_list = [
                [np.zeros((self.nx_l, self.N)) for k in range(self.admm_iters)]
                for i in range(self.n)
            ]
            x_plot_list = [
                [np.zeros((self.nx_l, self.N)) for k in range(self.admm_iters)]
                for i in range(self.n)
            ]
            x_back_plot_list = [
                [np.zeros((self.nx_l, self.N)) for k in range(self.admm_iters)]
                for i in range(self.n)
            ]

        for iter in range(self.admm_iters):
            if infeas_flag:
                break
            logger.debug(f"Greedy admm iter {iter}")
            # generate local sequences and choose one  - CHOICE: this can be done with vars from local output of ADMM
            # which may not have converged to consensus - therefore adding exploration OR a cooperative
            # dynamics rollout as before the loop
            if iter < self.switching_iters:
                for i in range(self.n):
                    if iter == 0:  # first iter we must used rolled out state
                        new_seqs = self.agents[i].eval_sequences(
                            x[i],
                            u[i],
                            [x_rout[j] for j in range(self.n) if self.Adj[i, j] == 1],
                        )
                        seqs[i] = new_seqs[0]  # use first by default for first iter
                        logger.debug(f"Agent {i} initial sez: {seqs[i]}")
                    else:
                        new_seqs = self.agents[i].eval_sequences(
                            x[i],
                            u[i],
                            xc[i],  # use local ADMM vars if not first iter
                        )

                        if seqs[i] in new_seqs:
                            new_seqs.remove(seqs[i])
                        if len(new_seqs) > 0:
                            logger.debug(
                                f"Agent {i} switched: {seqs[i]} to {new_seqs[0]}"
                            )
                            seqs[i] = new_seqs[0]  # for now choosing arbritrarily first

                            if ADMM_DEBUG_PLOT:
                                switch_plot_list[i].append(iter)
                    # set sequences
                    self.agents[i].set_sequence(seqs[i])

            # perform ADMM step
            action_list, sol_list, error_flag = self.admm_coordinator.solve_admm(state)

            if not error_flag:
                if ADMM_DEBUG_PLOT:
                    for i in range(self.n):
                        u_plot_list[i][iter] = np.asarray(sol_list[i].vals["u"])
                        z_plot_list[i][iter] = self.admm_coordinator.z[
                            i * self.nx_l : (i + 1) * self.nx_l, :
                        ]
                        x_plot_list[i][iter] = np.asarray(sol_list[i].vals["x"])
                        if i != self.n - 1:
                            x_back_plot_list[i][iter] = np.asarray(
                                sol_list[i].vals["x_c"][-2:, :]
                            )

                # extract the vars across the horizon from the ADMM sol for each agent
                for i in range(self.n):
                    u[i] = np.asarray(sol_list[i].vals["u"])
                    x_pred[i] = np.asarray(sol_list[i].vals["x"])
                    xc_out = np.asarray(sol_list[i].vals["x_c"])
                    xc_temp = []
                    for j in range(self.agents[i].num_neighbours):
                        xc_temp.append(xc_out[self.nx_l * j : self.nx_l * (j + 1), :])
                    xc[i] = xc_temp
            else:
                break

        if ADMM_DEBUG_PLOT:
            self.plot_admm_iters(
                u_plot_list,
                z_plot_list,
                x_plot_list,
                x_back_plot_list,
                switch_plot_list,
            )

        self.prev_sol = u
        self.prev_traj = x_pred
        return cs.DM(action_list), sol_list, error_flag, infeas_flag

    def dynamics_rollout(self, x: list[np.ndarray], u: list[np.ndarray]):
        """For a given state and u, rollout the agents' dynamics step by step."""
        x_temp = [np.zeros((self.nx_l, self.N)) for i in range(self.n)]

        for i in range(self.n):
            x_temp[i][:, [0]] = x[i]  # add the first known states to the temp
        for k in range(1, self.N):
            for i in range(self.n):
                xc_temp = []
                for j in range(self.n):
                    if self.Adj[i, j] == 1:
                        xc_temp.append(x_temp[j][:, [k - 1]])
                x_temp[i][:, [k]] = self.agents[i].next_state(
                    x_temp[i][:, [k - 1]], u[i][:, [k - 1]], xc_temp
                )
        return x_temp

    def plot_admm_iters(self, u_list, z_list, x_list, x_back_list, switch_list):
        t = 1

        _, u_axs = plt.subplots(self.n, 1, constrained_layout=True, sharex=True)
        _, res_axs = plt.subplots(self.n, 1, constrained_layout=True, sharex=True)
        for i in range(len(u_list)):
            u_axs[i].plot([u_list[i][k][:, t] for k in range(len(u_list[i]))])
            u_axs[i].plot(
                switch_list[i], [u_list[i][k][0, t] for k in switch_list[i]], "o"
            )
            u_axs[i].set_ylabel(f"u {i}")
            u_axs[i].set_xlabel("admm iter")
            # self.u_axs[i].set_ylim((-1, 1))

        t = 1
        _, axs = plt.subplots(1, 1, constrained_layout=True, sharex=True)
        axs.plot([x_list[2][k][0, t] for k in range(len(u_list[i]))])
        axs.plot([x_back_list[1][k][0, t] for k in range(len(u_list[i]))])
        # axs.plot([z_list[2][k][0, t] for k in range(len(u_list[i]))])
        # for i in range(len(z_list)):
        #   axs[i].plot([x_list[i][k][1, t] for k in range(len(u_list[i]))])
        #   if i != 0:
        #       axs[i].plot(
        #           [x_back_list[i - 1][k][1, t] for k in range(len(u_list[i]))]
        #       )

        for i in range(len(z_list)):
            # res_axs[i].plot(
            #    [x_list[i][k][1, t] for k in range(len(u_list[i]))]
            # )
            res_axs[i].plot(
                [x_list[i][k][1, t] - z_list[i][k][1, t] for k in range(len(u_list[i]))]
            )
            # res_axs[i].plot(
            #    switch_list[i],
            #    [x_list[i][k][1, t] - z_list[i][k][1, t] for k in switch_list[i]],
            #    "o",
            # )
            res_axs[i].set_ylabel(f"residual {i}")
            res_axs[i].set_xlabel("admm iter")
            # self.res_axs[i].set_ylim((-1, 1))
        plt.show()



from itertools import product

import numpy as np
from scipy.linalg import block_diag


def cent_from_dist(d_systems: list[dict], Adj: np.ndarray):
    """Creates a centralised representation of the distributed PWA system.
    PWA dynamics represented as: x+ = A[i]x + B[i]u + c[i] if S[i]x + R[i]u <= T.
    With state and control constraints Dx <= E, Fu <= G.

    Parameters
    ----------
    d_systems: List[dict]
        List of systems, each for an agent. Systems of form {S, R, T, A, B, c, D, E, F, G, [Aj]}.
    Adj: np.ndarray
        Adjacency matrix for system.
    """

    n = len(d_systems)  # num sub-systems
    nx_l = d_systems[0]["A"][0].shape[0]  # local x dim
    s_l = len(d_systems[0]["S"])  # num switching regions for each sub
    s = s_l**n  # num switching regions for centralised sys

    # each entry of the list contains a possible permutation of the PWA region indexes
    lst = [list(i) for i in product([j for j in range(s_l)], repeat=n)]

    S = []
    R = []
    T = []
    A = []
    B = []
    c = []

    # loop through every permutation
    for idxs in lst:
        S_tmp = []
        R_tmp = []
        T_tmp = []
        A_tmp = []
        B_tmp = []
        c_tmp = []
        for i in range(n):  # add the relevant region for each agent
            S_tmp.append(d_systems[i]["S"][idxs[i]])
            R_tmp.append(d_systems[i]["R"][idxs[i]])
            T_tmp.append(d_systems[i]["T"][idxs[i]])
            A_tmp.append(d_systems[i]["A"][idxs[i]])
            B_tmp.append(d_systems[i]["B"][idxs[i]])
            c_tmp.append(d_systems[i]["c"][idxs[i]])

        A_diag = block_diag(*A_tmp)
        # add coupling to A
        for i in range(n):
            coupling_idx = 0  # keep track of which Ac_i_j corresponds to Adj[i, j]
            for j in range(n):
                if Adj[i, j] == 1:
                    A_diag[
                        nx_l * i : nx_l * (i + 1), nx_l * j : nx_l * (j + 1)
                    ] = d_systems[i]["Ac"][idxs[i]][coupling_idx]
                    coupling_idx += 1

        S.append(block_diag(*S_tmp))
        R.append(block_diag(*R_tmp))
        T.append(np.vstack(T_tmp))
        A.append(A_diag)
        B.append(block_diag(*B_tmp))
        c.append(np.vstack(c_tmp))

    # state and control constraints which don't switch by region
    D = block_diag(*[d_systems[i]["D"] for i in range(n)])
    E = np.vstack([d_systems[i]["E"] for i in range(n)])
    F = block_diag(*[d_systems[i]["F"] for i in range(n)])
    G = np.vstack([d_systems[i]["G"] for i in range(n)])

    return {
        "S": S,
        "R": R,
        "T": T,
        "A": A,
        "B": B,
        "c": c,
        "D": D,
        "E": E,
        "F": F,
        "G": G,
    }



import casadi as cs
import numpy as np
from dmpcrl.mpc.mpc_admm import MpcAdmm


class MpcSwitching(MpcAdmm):
    """An admm based mpc for a pwa system. Dynamics and constrains switch every time step of the horizon.
    They are initialised as all zeros, to be set by the agent."""

    def set_dynamics(
        self,
        nx_l: int,
        nu_l: int,
        r: int,
        x: cs.SX,
        u: cs.SX,
        x_c_list: list[cs.SX],
    ) -> None:
        """Initialised the switching dynamics and constraints with the correct dimension and all zero values

        Parameters
        ----------
        nx_l: int
            Dimension of local state.
        nu_l: int
            Dimension of local control.
        r: int
            Number of rows in inequality that checks regions Sx + Ru <= T.
        x: cs.SX
            Local state symbolic var.
        u: cs.SX
            Local control symbolic var.
        x_c_list: List[cs.SX]
            List of coupled state symbolic vars."""

        num_neighbours = len(x_c_list)

        # create the params for switching dynamics and contraints
        A_list = []
        B_list = []
        c_list = []
        S_list = []
        R_list = []
        T_list = []
        Ac_list = []
        for k in range(self.horizon):
            A_list.append(self.parameter(f"A_{k}", (nx_l, nx_l)))
            B_list.append(self.parameter(f"B_{k}", (nx_l, nu_l)))
            c_list.append(self.parameter(f"c_{k}", (nx_l, 1)))
            S_list.append(self.parameter(f"S_{k}", (r, nx_l)))
            R_list.append(self.parameter(f"R_{k}", (r, nu_l)))
            T_list.append(self.parameter(f"T_{k}", (r, 1)))
            Ac_list.append([])
            for i in range(num_neighbours):
                Ac_list[k].append(self.parameter(f"Ac_{k}_{i}", (nx_l, nx_l)))

            # initialise as all zeros
            self.fixed_pars_init[f"A_{k}"] = np.zeros((nx_l, nx_l))
            self.fixed_pars_init[f"B_{k}"] = np.zeros((nx_l, nu_l))
            self.fixed_pars_init[f"c_{k}"] = np.zeros((nx_l, 1))
            self.fixed_pars_init[f"S_{k}"] = np.zeros((r, nx_l))
            self.fixed_pars_init[f"R_{k}"] = np.zeros((r, nu_l))
            self.fixed_pars_init[f"T_{k}"] = np.zeros((r, 1))
            for i in range(num_neighbours):
                self.fixed_pars_init[f"Ac_{k}_{i}"] = np.zeros((nx_l, nx_l))

        # dynamics and region constraints - added manually due to coupling
        for k in range(self.horizon):
            coup = cs.SX.zeros(nx_l, 1)
            for i in range(num_neighbours):  # get coupling expression
                coup += Ac_list[k][i] @ x_c_list[i][:, [k]]
            self.constraint(
                f"dynam_{k}",
                A_list[k] @ x[:, [k]] + B_list[k] @ u[:, [k]] + c_list[k] + coup,
                "==",
                x[:, [k + 1]],
            )
            self.constraint(
                f"region_{k}",
                S_list[k] @ x[:, [k]] + R_list[k] @ u[:, [k]],
                "<=",
                T_list[k],
            )
